#
#
#
#
#


import os
import unittest
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

from phlop.app.cmake import list_tests as get_cmake_tests
from phlop.os import env_sep
from phlop.proc import run
from phlop.reflection import classes_in_file
from phlop.sys import extend_sys_path

_LOG_DIR = Path(os.environ.get("PHLOP_LOG_DIR", os.getcwd()))

CMD_PREFIX = ""
CMD_POSTFIX = ""


@dataclass
class TestCase:
    cmd: str
    env: dict = field(default_factory=lambda: {})  # dict[str, str] # eventually
    working_dir: str = field(default_factory=lambda: None)
    log_file_path: str = field(default_factory=lambda: None)
    cores: int = field(default_factory=lambda: 1)

    def __post_init__(self):
        self.cmd = self.cmd.strip()


@dataclass
class TestBatch:
    tests: list
    cores: int


class DefaultTestCaseExtractor:
    def __call__(self, ctest_test):
        return [
            TestCase(
                cmd=ctest_test.cmd,
                env=ctest_test.env,
                working_dir=ctest_test.working_dir,
                log_file_path=_LOG_DIR
                / ".phlop"
                / f"{Path(ctest_test.working_dir).relative_to(_LOG_DIR)}",
            )
        ]


class GoogleTestCaseExtractor:
    def __call__(self, ctest_test):
        ...
        # not configured, assumed fast per file
        # print("GoogleTestCaseExtractor")
        # exec binary with `--gtest_list_tests` and see if it doesn't fail
        # p = run(
        #     ctest_test.cmd + " --gtest_list_tests --gtest_output=json:gtest.json",
        #     working_dir=ctest_test.working_dir,
        # )
        # print(p.stdout)

        return None


class PythonUnitTestCaseExtractor:
    def __call__(self, ctest_test):
        if "python3 " in ctest_test.cmd:  # hacky
            return load_py_test_cases_from_cmake(ctest_test)
        return None


EXTRACTORS = [
    PythonUnitTestCaseExtractor(),
    GoogleTestCaseExtractor(),
    DefaultTestCaseExtractor(),
]


def python3_default_test_cmd(clazz, test_id):
    return f"python3 -Oum {clazz.__module__} {clazz.__name__}.{test_id}"


def load_test_cases_in(
    classes, test_cmd_pre="", test_cmd_post="", test_cmd_fn=None, **kwargs
):
    test_cmd_fn = test_cmd_fn if test_cmd_fn else python3_default_test_cmd
    log_file_path = kwargs.pop("log_file_path", None)
    tests, loader = [], unittest.TestLoader()
    for test_class in classes:
        for suite in loader.loadTestsFromTestCase(test_class):
            cmd = test_cmd_fn(type(suite), suite._testMethodName)
            tests += [
                TestCase(
                    cmd=f"{test_cmd_pre} {cmd} {test_cmd_post}".strip(),
                    log_file_path=None
                    if not log_file_path
                    else f"{log_file_path}/{suite._testMethodName}",
                    **kwargs,
                )
            ]
    return tests


def load_py_test_cases_from_cmake(ctest_test):
    ppath = ctest_test.env.get("PYTHONPATH", "")
    bits = ctest_test.cmd.split(" ")
    idx = [i for i, x in enumerate(bits) if "python3" in x][0]
    prefix = " ".join(bits[:idx])
    with extend_sys_path([ctest_test.working_dir] + ppath.split(env_sep())):
        pyfile = bits[-1]
        return load_test_cases_in(
            classes_in_file(pyfile, unittest.TestCase, fail_on_import_error=True),
            env=ctest_test.env,
            working_dir=ctest_test.working_dir,
            log_file_path=_LOG_DIR
            / ".phlop"
            / f"{Path(ctest_test.working_dir).relative_to(_LOG_DIR)}",
            test_cmd_pre=CMD_PREFIX + prefix + CMD_POSTFIX,
        )


def determine_cores_for_test_case(test_case):
    try:
        if "mpirun -n" in test_case.cmd:
            bits = test_case.cmd.split(" ")
            idx = [i for i, x in enumerate(bits) if "mpirun" in x][0]
            test_case.cores = int(bits[idx + 2])
    except Exception as e:
        print("EXXXX", e)

    return test_case


def binless(test_case):
    if test_case.cmd.startswith("/usr/bin/"):
        test_case.cmd = test_case.cmd[9:]
    return test_case


MUTATORS = [determine_cores_for_test_case, binless]


def load_cmake_tests(cmake_dir, cores=1, test_cmd_pre="", test_cmd_post=""):
    cmake_tests = get_cmake_tests(cmake_dir)
    tests = []
    for cmake_test in cmake_tests:
        tests += [
            TestCase(
                cmd=test_cmd_pre + " ".join(cmake_test.command) + " " + test_cmd_post,
                env=cmake_test.env,
                working_dir=cmake_test.working_dir,
            )
        ]

    test_batches = {}

    def _add(test_cases):
        for test_case in test_cases:
            for mutator in MUTATORS:
                test_case = mutator(test_case)
            if test_case.cores not in test_batches:
                test_batches[test_case.cores] = []
            test_batches[test_case.cores].append(test_case)

    test_cases = []
    for test in tests:
        for extractor in EXTRACTORS:
            res = extractor(test)
            if res:
                _add(res)
                break

    return [TestBatch(v, k) for k, v in test_batches.items()]


@dataclass
class TestBatchesList:
    batch_list: List[TestBatch]


def deserialize(s):
    import codecs

    import dill

    return dill.loads(codecs.decode(s, "hex"))


def extract_load(cli_args):
    test_batches = {}

    file_paths = list(Path(cli_args.dir).glob(cli_args.load))

    if not file_paths:
        raise ValueError("No load files found")

    for path in file_paths:
        with open(path, "r") as file:
            batch_list = deserialize(file.read()).batch_list
            for batch in batch_list:
                for test_case in batch.tests:
                    if batch.cores not in test_batches:
                        test_batches[test_case.cores] = []
                    test_batches[test_case.cores].append(test_case)

    return [TestBatch(v, k) for k, v in test_batches.items()]



#
#
#
#
#


def decode_bytes(input, as_format="utf-8"):
    if isinstance(input, str):
        return input
    if isinstance(input, bytes):
        return input.decode(as_format, errors="ignore")
    raise ValueError(f"phlop.string::decode_bytes unknown input type: {type(input)}")



"""
  This script exists to minimize testing time by running all simulation/tests
    concurrently without needing to wait for any particular file or set of tests
"""

import json
import os
import unittest
from multiprocessing import cpu_count
from pathlib import Path

from phlop.reflection import classes_in_directory
from phlop.testing import parallel_processor as pp

N_CORES = int(os.environ["N_CORES"]) if "N_CORES" in os.environ else cpu_count()
PRINT = bool(json.loads(os.environ.get("PRINT", "false")))


def get_test_classes():
    return pp.load_test_cases_in(
        classes_in_directory(str(Path("tests")), unittest.TestCase)
    )


if __name__ == "__main__":
    pp.process(pp.TestBatch(get_test_classes(), 1), n_cores=N_CORES, print_only=PRINT)



#
#
#
#
#


import os
import sys
from contextlib import contextmanager


@contextmanager
def extend_sys_path(paths):
    if isinstance(paths, str):
        paths = [paths]
    old_path = sys.path[:]
    sys.path.extend(paths)
    try:
        yield
    finally:
        sys.path = old_path


@contextmanager
def extend_env(**environ):
    old_env = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_env)



#
#
#
#
#


import logging
import multiprocessing
import sys
import re
import unittest
from pathlib import Path

from phlop.dict import ValDict
from phlop.reflection import classes_in_directory
from phlop.testing import parallel_processor as pp

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def cli_args_parser():
    import argparse

    _help = ValDict(
        dir="Working directory",
        cmake="Enable cmake build config tests extraction",
        cores="Parallism core/thread count",
        print_only="Print only, no execution",
        prefix="Prepend string to execution string",
        postfix="Append string to execution string",
        dump="Dump discovered tests as YAML, no execution",
        load="Run tests exported from dump",
        regex="Filter out non-matching execution strings",
        logging="0=off, 1=on non zero exit code, 2=always",
    )

    parser = argparse.ArgumentParser()
    parser.add_argument("--cmake", action="store_true", default=False, help=_help.cmake)
    parser.add_argument("-c", "--cores", type=int, default=1, help=_help.cores)
    parser.add_argument("-d", "--dir", default=".", help=_help.dir)
    parser.add_argument("-f", "--filter", type=str, default="", help="")
    parser.add_argument(
        "-p", "--print_only", action="store_true", default=False, help=_help.print_only
    )
    parser.add_argument("--prefix", default="", help=_help.prefix)
    parser.add_argument("--postfix", default="", help=_help.postfix)

    parser.add_argument(
        "--dump", default=None, action="store", nargs="?", help=_help.dump
    )
    parser.add_argument("--load", default=None, help=_help.load)
    parser.add_argument("-r", "--regex", default=None, help=_help.regex)
    parser.add_argument("--logging", type=int, default=1, help=_help.logging)

    return parser


def verify_cli_args(cli_args):
    if cli_args.cores == "a" or cli_args.cores == "all":
        cli_args.cores = multiprocessing.cpu_count()
    cli_args.cores = int(cli_args.cores)
    if not Path(cli_args.dir).exists():
        raise RuntimeError(
            "phlop.run.test_cases error: directory provided does not exist"
        )
    pp.LoggingMode(cli_args.logging)  # check convertible
    sys.argv = [sys.argv[0]]  # drop everything!
    return cli_args


def get_test_cases(cli_args):
    if cli_args.cmake:
        return pp.load_cmake_tests(
            cli_args.dir, test_cmd_pre=cli_args.prefix, test_cmd_post=cli_args.postfix
        )
    return [
        pp.TestBatch(
            pp.load_test_cases_in(
                classes_in_directory(cli_args.dir, unittest.TestCase),
                test_cmd_pre=cli_args.prefix,
                test_cmd_post=cli_args.postfix,
            ),
            1,
        )
    ]


def dump_batches(cli_args):
    import codecs

    import dill

    with open(cli_args.dump, "w") as f:
        f.write(
            codecs.encode(
                dill.dumps(pp.TestBatchesList(batch_list=get_test_cases(cli_args))),
                "hex",
            ).decode("utf8")
        )


def filter_out_regex_fails(cli_args, test_batches):
    if not cli_args.regex:
        return test_batches
    try:
        pattern = re.compile(cli_args.regex)
        is_valid = True
        op = lambda x: pattern.search(x)
    except re.error:
        print("regex invalid, resorting to 'str in str' approach")
        is_valid = False
        op = lambda x: regex in x
    filtered = {tb.cores: [] for tb in test_batches}
    for tb in test_batches:
        for test in tb.tests:
            if op(test.cmd):
                filtered[tb.cores].append(test)
    return [pp.TestBatch(v, k) for k, v in filtered.items() if v]


def noLog(test_batches):
    for tb in test_batches:
        for test in tb.tests:
            test.log_file_path = None
    return test_batches


def log(test_batches):
    # synchronously make all directories before executions
    for tb in test_batches:
        for test in tb.tests:
            if test.log_file_path:
                Path(test.log_file_path).parent.mkdir(parents=True, exist_ok=True)
    return test_batches


def main():
    parser = cli_args_parser()
    cli_args = verify_cli_args(parser.parse_args())
    try:
        if cli_args.dump and cli_args.load:
            raise ValueError("Cannot use 'dump' and 'load' options simultaneously.")

        if cli_args.dump:
            dump_batches(cli_args)
            return

        test_batches = (
            pp.extract_load(cli_args) if cli_args.load else get_test_cases(cli_args)
        )
        test_batches = filter_out_regex_fails(cli_args, test_batches)

        if cli_args.logging == 0:
            test_batches = noLog(test_batches)
        else:
            test_batches = log(test_batches)

        pp.process(
            test_batches,
            n_cores=cli_args.cores,
            print_only=cli_args.print_only,
            logging=cli_args.logging,
        )

    except pp.TestCaseFailure as e:
        sys.exit(1)
    except (Exception, SystemExit) as e:
        logger.exception(e)
        parser.print_help()
    except:
        e = sys.exc_info()[0]
        print(f"Error: Unknown Error {e}")


if __name__ == "__main__":
    main()



#
#
#
#
#


import csv
import os

from phlop.proc import run, run_mp


def version():
    # validated on perf version: 5.19
    proc = run("perf -v", shell=True, capture_output=True).out()
    if " " not in proc or "." not in proc:
        raise ValueError("Unparsable result from 'perf -v'")
    return [int(digit) for digit in proc.split(" ")[-1].split(".")]


def check(force_kernel_space=False):
    """perf can require some system config / read the error if thrown"""
    kernel_space_opt = "a" if force_kernel_space else ""
    run(
        f"perf stat -{kernel_space_opt}d sleep 1",
        shell=True,
        capture_output=True,
        check=True,
    )
    record("ls", [], "/tmp/perf_record_check.dat")


def parse_key(key, force_kernel_space):
    user_space_postfix = ":u"
    if key.endswith(user_space_postfix):
        if force_kernel_space:
            raise RuntimeError(f"Userspace event found {key}")
        return key[: -len(user_space_postfix)]
    return key


def parse_stat_csv(file, force_kernel_space=False):
    comments_lines = 2  # validate
    row_val_idx, row_id_idx = 0, 2
    with open(file, newline="") as csvfile:
        [next(csvfile) for i in range(comments_lines)]  # skip headers
        return {
            parse_key(row[row_id_idx], force_kernel_space): row[row_val_idx]
            for row in csv.reader(csvfile, delimiter=",")
        }


# https://perf.wiki.kernel.org/index.php/Tutorial
# http://www.brendangregg.com/perf.html
# or run "perf list"
def events_str(events):
    if len(events) == 0:
        return ""
    return f"-e {events if isinstance(events, str) else ','.join(events)}"


def out_str(output_file):
    return "" if output_file is None else f"-o {os.path.relpath(output_file)}"


def stat_cmd(exe, events, output_file, options=""):
    return f"perf stat -x , {options} {out_str(output_file)} {events_str(events)} {exe}"


def stat(exe, events, output_file=None):
    return run(stat_cmd(exe, events, output_file), check=True)


def record_cmd(exe, events, output_file, options=""):
    return f"perf record {options} {out_str(output_file)} {events_str(events)} {exe}"


def record(exe, events, output_file=None):
    return run(record_cmd(exe, events, output_file), check=True)


def stat_mp(exe, events, output_files):
    return run_mp([stat_cmd(exe, events, out) for out in output_files])



#
#
#
#
#


import time
from enum import Enum
from multiprocessing import Process, Queue, cpu_count

from phlop.proc import ProcessNonZeroExitCode, run
from phlop.testing.test_cases import *


class TestCaseFailure(Exception):
    ...


class LoggingMode(Enum):
    OFF = 0
    ON_NON_ZERO_EXIT_CODE = 1
    ALWAYS = 2


class CallableTest:
    def __init__(self, batch_index, test_case, logging):
        self.batch_index = batch_index
        self.test_case = test_case
        self.run = None
        self.logging = logging

    def __call__(self, **kwargs):
        self.run = run(
            self.test_case.cmd.split(),
            shell=False,
            capture_output=True,
            check=True,
            print_cmd=False,
            env=self.test_case.env,
            working_dir=self.test_case.working_dir,
            log_file_path=self.test_case.log_file_path,
            logging=self.logging,
        )
        if self.run.stderr and self.run.exitcode != 0:
            print(self.run.stderr)
        return self


class CoreCount:
    def __init__(self, cores_avail):
        self.cores_avail = cores_avail
        self.procs = []
        self.fin = []


def runner(runnable, queue):
    queue.put(runnable())


def print_tests(batches):
    for batch in batches:
        for test in batch.tests:
            print(test.cmd)


def process(
    batches, n_cores=None, print_only=False, fail_fast=False, logging: LoggingMode = 1
):
    if not isinstance(batches, list):
        batches = [batches]
    if sum([len(t.tests) for t in batches]) == 0:
        return  # nothing to do

    if print_only:
        print_tests(batches)
        return

    n_cores = n_cores if n_cores else cpu_count()

    cc = CoreCount(n_cores)
    assert cc.cores_avail >= max([batch.cores for batch in batches])
    cc.procs = [[] for batch in batches]
    cc.fin = [0 for batch in batches]
    pqueue = Queue()

    def launch_tests():
        for batch_index, batch in enumerate(batches):
            offset = len(cc.procs[batch_index])
            for test_index, test in enumerate(batch.tests[offset:]):
                test_index += offset
                if batch.cores <= cc.cores_avail:
                    test = CallableTest(
                        batch_index, batches[batch_index].tests[test_index], logging
                    )
                    cc.cores_avail -= batch.cores
                    cc.procs[batch_index] += [
                        Process(
                            target=runner,
                            args=(test, (pqueue)),
                        )
                    ]
                    cc.procs[batch_index][-1].daemon = True
                    cc.procs[batch_index][-1].start()

    def finished():
        b = True
        for batch_index, batch in enumerate(batches):
            b &= cc.fin[batch_index] == len(batch.tests)
        return b

    def waiter(queue):
        fail = 0
        while True:
            proc = queue.get()
            time.sleep(0.01)  # don't throttle!
            if isinstance(proc, CallableTest):
                status = "finished" if proc.run.exitcode == 0 else "FAILED"
                fail += proc.run.exitcode
                if fail_fast and fail > 0:
                    raise TestCaseFailure("Some tests have failed")
                print(
                    proc.test_case.cmd, f"{status} in {proc.run.run_time:.2f} seconds"
                )
                cc.cores_avail += batches[proc.batch_index].cores
                cc.fin[proc.batch_index] += 1
                if finished():
                    if fail > 0:
                        raise TestCaseFailure("Some tests have failed")
                    break
                launch_tests()

    launch_tests()
    waiter(pqueue)



#
#
#
#
#


import json
from dataclasses import dataclass, field

from phlop.proc import run
from phlop.string import decode_bytes


def version():
    pass


def make_config_str(path, cxx_flags=None, use_ninja=False, use_ccache=False, extra=""):
    cxx_flags = "" if cxx_flags is None else f'-DCMAKE_CXX_FLAGS="{cxx_flags}"'
    ccache = "" if use_ccache is False else "-DCMAKE_CXX_COMPILER_LAUNCHER=ccache"
    ninja = "" if not use_ninja else "-G Ninja"

    return f"cmake {path} {cxx_flags} {ninja} {ccache} {extra}"


def config(path, cxx_flags=None, use_ninja=False, use_ccache=False, extra=""):
    cmd = make_config_str(path, cxx_flags, use_ninja, extra)
    run(cmd, capture_output=False)


def build(use_ninja=False, threads=1):
    run("ninja" if use_ninja else f"make -j{threads}", capture_output=False)


@dataclass
class CTest_test:
    backtrace: int  # care?
    command: list  # list[str] # eventually
    # [
    #   "/opt/py/py/bin/python3",
    #   "-u",
    #   "test_particles_advance_2d.py"
    # ],
    name: str  # "py3_advance-2d-particles",
    properties: list = field(default_factory=lambda: [{}])  # list[dict] # eventually

    env: dict = field(default_factory=lambda: {})  # dict[str, str] # eventually
    working_dir: str = field(default_factory=lambda: None)

    def __post_init__(self):
        for p in self.properties:
            if p["name"] == "ENVIRONMENT":
                for item in p["value"]:
                    bits = item.split("=")
                    self.env.update({bits[0]: "=".join(bits[1:])})
            elif p["name"] == "WORKING_DIRECTORY":
                self.working_dir = p["value"]

    # [
    #   {
    #     "name" : "ENVIRONMENT",
    #     "value" :
    #     [
    #       "PYTHONPATH=/home/p/git/phare/master/build:/home/p/git/phare/master/pyphare",
    #       "ASAN_OPTIONS=detect_leaks=0"
    #     ]
    #   },
    #   {
    #     "name" : "WORKING_DIRECTORY",
    #     "value" : "/home/p/git/phare/master/build/tests/simulator/advance"
    #   }
    # ]


def list_tests(build_dir=None):
    cmd = "".join(
        [
            s
            for s in [
                "ctest ",
                f"--test-dir {build_dir} " if build_dir else "",
                "--show-only=json-v1",
            ]
            if s
        ]
    )
    return [
        CTest_test(**test)
        for test in json.loads(decode_bytes(run(cmd, capture_output=True).stdout))[
            "tests"
        ]
    ]


def test_cmd(test, verbose=False):
    cmd = f"ctest -R {test}"
    if verbose:
        cmd = f"{cmd} -V"
    return cmd


if __name__ == "__main__":
    list_tests("build")



from setuptools import find_packages, setup

setup(
    name="phlop",
    version="0.0.14",
    cmdclass={},
    classifiers=[],
    include_package_data=True,
    packages=find_packages(exclude=["lib/", "tests/"]),
)



#
#
#
#
#


import atexit
import subprocess

from phlop.proc import run
from phlop.string import decode_bytes


def current_branch():
    return decode_bytes(run("git branch --show-current").stdout)


def hashes(N=1):
    return decode_bytes(run(f"git log -{N} --pretty=format:%h").stdout).splitlines()


def log(N=10, use_short=False):
    short = '--pretty=format:"%h%x09%an%x09%ad%x09%s"' if use_short else ""
    return decode_bytes(run(f"git log -{N} {short}").stdout)


def branch_exists(branch):
    try:
        run(f"git show-branch {branch}", check=True)
    except subprocess.CalledProcessError as e:
        return False  # exit failure means branch does not exist
    return True


def checkout(branch, create=False):
    if create and not branch_exists(branch):
        run(f"git checkout -b {branch}", check=True)
    else:
        run(f"git checkout {branch}", check=True)


def git_branch_reset_at_exit():
    current_git_branch = current_branch()

    @atexit.register
    def _reset_():
        checkout(current_git_branch)
